// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import java.util.*;
import org.xwt.util.*;
import org.xwt.js.*;

/**
 *  A background thread. All threads created with <tt>xwt.thread</tt>
 *  are instances of this class. ThreadMessage objects can be enqueued
 *  in MessageQueue.
 *
 *  INVARIANT: only one thread can be executing JavaScript scripts or
 *             accessing Box instances at any given time. For
 *             performance reasons, no locks are employed to enforce
 *             this invariant.
 */
public class ThreadMessage extends Thread implements Message {
    
    private volatile static int threadcount = 0;

    /** the JavaScript function that we are executing */
    volatile JS.Callable f;

    /** the ThreadMessage thread blocks on this before executing any JavaScript */
    Semaphore go = new Semaphore();

    /** The MessageQueue (main) thread blocks on this while the ThreadMessage thread is running JavaScript code */
    Semaphore done = new Semaphore();

    /** used to pool ThreadMessages that are not in use */
    private static Queue spare = new Queue(50);

    /** queue of functions waiting to be spawned; used if threadcount == Platform.maxThreads() */
    private static Queue waiting = new Queue(50);

    private ThreadMessage() { start(); }
    private static Object[] emptyobj = new Object[] { };

    /** creates a new thread to execute function <tt>f</tt> */
    public static synchronized void newthread(JS.Callable f) {
        ThreadMessage ret = (ThreadMessage)spare.remove(false);
        if (ret == null) {
            if (threadcount < Platform.maxThreads()) ret = new ThreadMessage();
            else {
                waiting.append(f);
                return;
            }
        }
        ret.f = f;
        MessageQueue.add(ret);
    }

    /** attempts to put this thread into the background to perform a blocking operation; returns false if unable to do so */
    public static boolean suspendThread() {
        // put ourselves in the background
        Thread thread = Thread.currentThread();
        if (!(thread instanceof ThreadMessage)) {
            if (Log.on) Log.logJS(ThreadMessage.class, "attempt to perform background-only operation in a foreground thread");
            return false;
        }
        ThreadMessage mythread = (ThreadMessage)thread;
        mythread.setPriority(Thread.MIN_PRIORITY);
        mythread.done.release();
        return true;
    }

    /** re-enqueues this thread */
    public static void resumeThread() {
        ThreadMessage mythread = (ThreadMessage)Thread.currentThread();
        MessageQueue.add(mythread);
        mythread.setPriority(Thread.NORM_PRIORITY);
        mythread.go.block();        
    }
    
    public void run() {
        try {
            threadcount++;
            while (true) {
                try {
                    go.block();
                    f.call(new JS.Array());
                } catch (JS.Exn e) {
                    if (Log.on) Log.log(this, "WARNING: uncaught ecmascript exception: " + e);
                }
                done.release();
                synchronized(waiting) {
                    if (waiting.size() > 0) {
                        f = (JS.Callable)waiting.remove(false);
                        MessageQueue.add(this);
                    } else if (spare.size() < 10) {
                        spare.append(this);
                    } else {
                        threadcount--;
                        return;
                    }
                }
            }
        } catch (Throwable t) {
            if (Log.on) Log.log(this, "caught throwable at thread entry point; this should never happen");
            if (Log.on) Log.log(this, t);
            if (Log.on) Log.log(this, "reaping thread");
            return;
        }
    }

    /** this is invoked in the MessageQueue thread */
    public void perform() {
        go.release();
        done.block();
    }

}
